@@ -33,6 +33,7 @@ module Config.Schema.Load.Error
   -- * Summaries
   , describeSpec
   , describeValue
+  , simplifyValueSpecMismatch
   ) where
 
 import           Control.Exception
@@ -47,6 +48,7 @@ import           Text.PrettyPrint
                      punctuate, comma, int, colon, hcat)
 
 import           Config
+import           Config.Macro (FilePosition(..))
 import           Config.Schema.Types
 
 #if !MIN_VERSION_base(4,11,0)
@@ -54,6 +56,8 @@ import           Data.Monoid ((<>))
 #endif
 
 -- | Newtype wrapper for schema load errors.
+--
+-- @since 1.2.0.0
 data ValueSpecMismatch p =
   -- | Problem value and list of specification failures
   ValueSpecMismatch p Text (NonEmpty (PrimMismatch p))
@@ -62,6 +66,8 @@ data ValueSpecMismatch p =
 -- | Type for errors that can be encountered while decoding a value according
 -- to a specification. The error includes a key path indicating where in
 -- the configuration file the error occurred.
+--
+-- @since 1.2.0.0
 data PrimMismatch p =
   -- | spec description and problem
   PrimMismatch Text (Problem p)
@@ -69,6 +75,8 @@ data PrimMismatch p =
 
 
 -- | Problems that can be encountered when matching a 'Value' against a 'ValueSpec'.
+--
+-- @since 1.2.0.0
 data Problem p
   = MissingSection Text                          -- ^ missing section name
   | UnusedSections (NonEmpty Text)               -- ^ unused section names
@@ -81,6 +89,8 @@ data Problem p
   deriving Show
 
 -- | Describe outermost shape of a 'PrimValueSpec'
+--
+-- @since 1.2.0.0
 describeSpec :: PrimValueSpec a -> Text
 describeSpec TextSpec                   = "text"
 describeSpec NumberSpec                 = "number"
@@ -122,17 +132,15 @@ removeTypeMismatch1 (ValueSpecMismatch p v xs)
   = ValueSpecMismatch p v xs'
 removeTypeMismatch1 v = v
 
+-- | Returns 'True' for schema mismatches where the value type doesn't
+-- match.
 isTypeMismatch :: PrimMismatch p -> Bool
 isTypeMismatch (PrimMismatch _ prob) =
   case prob of
-    WrongAtom -> True
-    TypeMismatch -> True
-    NestedProblem x -> go x
-    SubkeyProblem _ x -> go x
-    ListElementProblem _ x -> go x
-    _ -> False
-  where
-    go (ValueSpecMismatch _ _ xs) = all isTypeMismatch xs
+    WrongAtom                                -> True
+    TypeMismatch                             -> True
+    NestedProblem (ValueSpecMismatch _ _ xs) -> all isTypeMismatch xs
+    _                                        -> False
 
 -- | Single-step rewrite that removes mismatches with only a single,
 -- nested mismatch below them.
@@ -151,6 +159,8 @@ focusMismatch1 x@(ValueSpecMismatch _ _ prims)
 -- | Pretty-printer for 'ValueSpecMismatch' showing the position
 -- and type of value that failed to match along with details about
 -- each specification that it didn't match.
+--
+-- @since 1.2.0.0
 prettyValueSpecMismatch :: ErrorAnnotation p => ValueSpecMismatch p -> Doc
 prettyValueSpecMismatch (ValueSpecMismatch p v es) =
   heading $+$ errors
@@ -162,14 +172,27 @@ prettyValueSpecMismatch (ValueSpecMismatch p v es) =
 -- | Pretty-printer for 'PrimMismatch' showing a summary of the primitive
 -- specification that didn't match followed by a more detailed error when
 -- appropriate.
+--
+-- @since 1.2.0.0
 prettyPrimMismatch :: ErrorAnnotation p => PrimMismatch p -> Doc
 prettyPrimMismatch (PrimMismatch spec problem) =
   case prettyProblem problem of
     (summary, detail) ->
-      (text "*" <+> text (Text.unpack spec) <+> summary) $+$ nest 4 detail
+      (text "* expected" <+> text (Text.unpack spec) <+> summary) $+$ nest 4 detail
+
+-- | Simplify a 'ValueSpecMismatch' by collapsing long nested error
+-- cases and by assuming that if a type matched that the other mismatched
+-- type alternatives are uninteresting. This is used in the implementation
+-- of 'displayException'.
+--
+-- @since 1.2.1.0
+simplifyValueSpecMismatch :: ValueSpecMismatch p -> ValueSpecMismatch p
+simplifyValueSpecMismatch = rewriteMismatch (focusMismatch1 . removeTypeMismatch1)
 
 -- | Pretty-printer for 'Problem' that generates a summary line
 -- as well as a detailed description (depending on the error)
+--
+-- @since 1.2.0.0
 prettyProblem ::
   ErrorAnnotation p =>
   Problem p ->
@@ -203,17 +226,29 @@ prettyProblem p =
       , prettyValueSpecMismatch e)
 
 -- | Class for rendering position annotations within the 'prettyValueSpecMismatch'
+--
+-- @since 1.2.0.0
 class (Typeable a, Show a) => ErrorAnnotation a where
   displayAnnotation :: a -> Doc
 
 -- | Renders a 'Position' as @line:column:@
+--
+-- @since 1.2.0.0
 instance ErrorAnnotation Position where
   displayAnnotation pos = hcat [int (posLine pos), colon, int (posColumn pos), colon]
 
+instance ErrorAnnotation FilePosition where
+  displayAnnotation (FilePosition path pos) = hcat [text path, colon, int (posLine pos), colon, int (posColumn pos), colon]
+
 -- | Renders as an empty document
+--
+-- @since 1.2.0.0
 instance ErrorAnnotation () where
   displayAnnotation _ = empty
 
 -- | 'displayException' implemented with 'prettyValueSpecMismatch'
+-- and 'simplifyValueSpecMismatch'.
+--
+-- @since 1.2.0.0
 instance ErrorAnnotation p => Exception (ValueSpecMismatch p) where
-  displayException = show . prettyValueSpecMismatch . rewriteMismatch (focusMismatch1 . removeTypeMismatch1)
+  displayException = show . prettyValueSpecMismatch . simplifyValueSpecMismatch
